'''
[Neural Style Transfer]
Gatys, Leon A, et al. Image Style Transfer Using Convolutional Neural Networks. CVPR 2016.
Code reference: https://github.com/leongatys/PytorchNeuralStyleTransfer

[运行]
python Practice_24_NeuralStyleTransfer.py

[环境要求]
- Pytorch
- torchvision
- opencv
- matplotlib
- pillow
- tqdm

[模型]
在本文件加的checkpoint目录下下载以下模型
wget -c --no-check-certificate https://bethgelab.org/media/uploads/pytorch_models/vgg_conv.pth

[数据]
在本文件夹的data/styletransfer/目录下下载以下数据用作测试
wget -O ./data/styletransfer/Tuebingen_Neckarfront.jpg "https://raw.githubusercontent.com/leongatys/PytorchNeuralStyleTransfer/master/Images/Tuebingen_Neckarfront.jpg"
wget -O ./data/styletransfer/vangogh_starry_night.jpg "https://raw.githubusercontent.com/leongatys/PytorchNeuralStyleTransfer/master/Images/vangogh_starry_night.jpg"
wget -O ./data/styletransfer/brick.jpg "https://raw.githubusercontent.com/Yijunmaverick/UniversalStyleTransfer/master/input/style/brick.jpg"
wget -O ./data/styletransfer/seated-nude.jpg "https://raw.githubusercontent.com/Yijunmaverick/UniversalStyleTransfer/master/input/style/seated-nude.jpg"
wget -O ./data/styletransfer/tiger.jpg "https://raw.githubusercontent.com/Yijunmaverick/UniversalStyleTransfer/master/input/style/088.jpg"
'''

import os
from PIL import Image
import matplotlib.pyplot as plt

import torchvision.transforms as transforms
import torchvision.models as models

import numpy as np
from datetime import datetime
import time
import logging
import random
import cv2
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.autograd as autograd
from torch.autograd import Variable
from torch.nn import Linear, Conv2d, BatchNorm2d, LayerNorm, LeakyReLU, ConvTranspose2d, ReLU, Tanh, InstanceNorm2d
from torch.nn import ReflectionPad2d, ReplicationPad2d

from torch.utils.data import DataLoader
import torchvision.datasets as dset
import torchvision
import itertools
from tqdm import tqdm


'''辅助函数''' 
# 保存图像
def save_image(img, filename):
    tmp = ((img.detach().numpy().transpose(1, 2, 0) + 1.0) * 127.5).astype(np.uint8)
    cv2.imwrite(filename, cv2.cvtColor(tmp, cv2.COLOR_RGB2BGR))
    
# 图像的预处理和后处理函数
img_size = 512 
prep = transforms.Compose([transforms.Resize(img_size),
                           transforms.ToTensor(),
                           transforms.Lambda(lambda x: x[torch.LongTensor([2,1,0])]), 
                           transforms.Normalize(mean=[0.40760392, 0.45795686, 0.48501961], 
                                                std=[1,1,1]),
                           transforms.Lambda(lambda x: x.mul_(255)),
                          ])
postpa = transforms.Compose([transforms.Lambda(lambda x: x.mul_(1./255)),
                           transforms.Normalize(mean=[-0.40760392, -0.45795686, -0.48501961], 
                                                std=[1,1,1]),
                           transforms.Lambda(lambda x: x[torch.LongTensor([2,1,0])]), 
                           ])

# 将图像的范围限制到[-1,1]
def postp(tensor): 
    t = postpa(tensor) * 2 - 1
    t[t>1] = 1    
    t[t<-1] = -1
    return t


'''损失函数定义'''
# 计算Gram矩阵
class GramMatrix(nn.Module):
    def forward(self, input):
        b,c,h,w = input.size()
        F = input.view(b, c, h*w)
        G = torch.bmm(F, F.transpose(1,2)) 
        G.div_(h*w)
        return G

# 计算风格损失
# # input为VGG特征，target为目标风格特征，首先将input转换为Gram矩阵，再计算其与target的均方误差
class GramMSELoss(nn.Module):
    def forward(self, input, target):
        out = nn.MSELoss()(GramMatrix()(input), target)
        return(out)

# 定义风格特征的层数    
style_layers = ['r11','r21','r31','r41', 'r51'] 
# 定义内容特征的层数
content_layers = ['r42']
# 整合所有计算损失函数的类
loss_layers = style_layers + content_layers
loss_fns = [GramMSELoss()] * len(style_layers) + [nn.MSELoss()] * len(content_layers)
if torch.cuda.is_available():
    loss_fns = [loss_fn.cuda() for loss_fn in loss_fns]
    
# 不同层风格损失和内容损失的权重
style_weights = [1e3/n**2 for n in [64,128,256,512,512]]
content_weights = [1e0]
weights = style_weights + content_weights

    
'''网络定义'''
# 定义VGG网络模型
class VGG(nn.Module):
    def __init__(self, pool='max'):
        super(VGG, self).__init__()
        #vgg modules
        self.conv1_1 = nn.Conv2d(3, 64, kernel_size=3, padding=1)
        self.conv1_2 = nn.Conv2d(64, 64, kernel_size=3, padding=1)
        self.conv2_1 = nn.Conv2d(64, 128, kernel_size=3, padding=1)
        self.conv2_2 = nn.Conv2d(128, 128, kernel_size=3, padding=1)
        self.conv3_1 = nn.Conv2d(128, 256, kernel_size=3, padding=1)
        self.conv3_2 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.conv3_3 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.conv3_4 = nn.Conv2d(256, 256, kernel_size=3, padding=1)
        self.conv4_1 = nn.Conv2d(256, 512, kernel_size=3, padding=1)
        self.conv4_2 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv4_3 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv4_4 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_1 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_2 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_3 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        self.conv5_4 = nn.Conv2d(512, 512, kernel_size=3, padding=1)
        if pool == 'max':
            self.pool1 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool2 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool3 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool4 = nn.MaxPool2d(kernel_size=2, stride=2)
            self.pool5 = nn.MaxPool2d(kernel_size=2, stride=2)
        elif pool == 'avg':
            self.pool1 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool2 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool3 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool4 = nn.AvgPool2d(kernel_size=2, stride=2)
            self.pool5 = nn.AvgPool2d(kernel_size=2, stride=2)
            
    # out_keys为指定的层数，根据out_keys提取x经过VGG后对应层数的特征        
    def forward(self, x, out_keys):
        out = {}
        out['r11'] = F.relu(self.conv1_1(x))
        out['r12'] = F.relu(self.conv1_2(out['r11']))
        out['p1'] = self.pool1(out['r12'])
        out['r21'] = F.relu(self.conv2_1(out['p1']))
        out['r22'] = F.relu(self.conv2_2(out['r21']))
        out['p2'] = self.pool2(out['r22'])
        out['r31'] = F.relu(self.conv3_1(out['p2']))
        out['r32'] = F.relu(self.conv3_2(out['r31']))
        out['r33'] = F.relu(self.conv3_3(out['r32']))
        out['r34'] = F.relu(self.conv3_4(out['r33']))
        out['p3'] = self.pool3(out['r34'])
        out['r41'] = F.relu(self.conv4_1(out['p3']))
        out['r42'] = F.relu(self.conv4_2(out['r41']))
        out['r43'] = F.relu(self.conv4_3(out['r42']))
        out['r44'] = F.relu(self.conv4_4(out['r43']))
        out['p4'] = self.pool4(out['r44'])
        out['r51'] = F.relu(self.conv5_1(out['p4']))
        out['r52'] = F.relu(self.conv5_2(out['r51']))
        out['r53'] = F.relu(self.conv5_3(out['r52']))
        out['r54'] = F.relu(self.conv5_4(out['r53']))
        out['p5'] = self.pool5(out['r54'])
        return [out[key] for key in out_keys]


'''构建和加载模型'''
# 构建VGG网络实例，并加载参数vgg_conv.pth
vgg = VGG()
vgg.load_state_dict(torch.load('./checkpoint/vgg_conv.pth'))
# 不训练网络，因此网络不需要计算梯度
for param in vgg.parameters():
    param.requires_grad = False
if torch.cuda.is_available():
    vgg.cuda()
print(vgg)

    
'''数据准备''' 
image_dir = './data/styletransfer/'
img_dirs = [image_dir, image_dir]
# 通过改变下面的文件名更改风格图像和内容图像
img_names = ['vangogh_starry_night.jpg', 'Tuebingen_Neckarfront.jpg']
imgs = [Image.open(img_dirs[i] + name) for i,name in enumerate(img_names)]
imgs_torch = [prep(img) for img in imgs]
if torch.cuda.is_available():
    imgs_torch = [Variable(img.unsqueeze(0).cuda()) for img in imgs_torch]
else:
    imgs_torch = [Variable(img.unsqueeze(0)) for img in imgs_torch]
# 风格图像和内容图像分别为style_image和content_image
style_image, content_image = imgs_torch

# 图像style_image的风格特征
style_targets = [GramMatrix()(A).detach() for A in vgg(style_image, style_layers)]
# 图像content_image的内容特征
content_targets = [A.detach() for A in vgg(content_image, content_layers)]
targets = style_targets + content_targets


'''网络训练'''
# 目标图像直接使用内容图像初始化
opt_img = Variable(content_image.data.clone(), requires_grad=True)
max_iter = 500  # 反向传播迭代次数
show_iter = 50  # 每50次迭代打印一次当前损失
optimizer = optim.LBFGS([opt_img]); # 优化器选用L-BFGS
n_iter=[0] # 当前迭代次数

out_img = postp(opt_img.data[0].cpu().squeeze())
save_image(out_img, 'NeuralStyleTransfer_%s.jpg'%(n_iter[0]))

while n_iter[0] <= max_iter:
    
    # closure函数定义了一次完整的迭代，包含损失函数计算及梯度反传
    def closure():
        optimizer.zero_grad()
        # 提取目标图像的VGG特征
        out = vgg(opt_img, loss_layers)
        # 计算每层的损失函数，及总的损失函数 
        # 权重weight，损失loss_fns和特征targets都已经整合了风格和内容两方面
        layer_losses = [weights[a] * loss_fns[a](A, targets[a]) for a,A in enumerate(out)]
        loss = sum(layer_losses)
        loss.backward()
        n_iter[0]+=1
        # 打印当前损失
        if n_iter[0]%show_iter == (show_iter-1):
            print('Iteration: %d, loss: %f'%(n_iter[0]+1, loss.data.mean()))
            out_img = postp(opt_img.data[0].cpu().squeeze())
            save_image(out_img, 'NeuralStyleTransfer_%s.jpg'%(n_iter[0]))
        return loss
    
    # 使用optimizer.step对目标图像进行一次更新
    optimizer.step(closure) 

print('Done!')